import numpy as np
from scipy.linalg import lstsq
from sklearn.svm import LinearSVC

def normalize(f):
    def _f(*args, **kwargs):
        w = f(*args, **kwargs)
        norm = np.linalg.norm(w)
        if norm != 0:
            return w / norm
        return w
    return _f

@normalize
def split_svm(A, B, svm_args=[]):
    X = np.concatenate((A, B), axis=0)
    y = np.array([0]*len(A)+[1]*len(B))
    return LinearSVC(*svm_args).fit(X, y).coef_.flatten()

def gen_split_f(f):
    @normalize
    def split(A, B):
        return f(B)-f(A)
    return split
split_median = gen_split_f(lambda v: np.median(v, axis=0))
split_mean = gen_split_f(lambda v: np.mean(v, axis=0))

@normalize
def split_dist(A, B):
    """At least 3 points between A and B
    
    Maximizes the distance between the projected A point cloud and the projected B point
    cloud, normalizing by the cumulative width of those point clouds (minimizing
    overlap for most real-world probability distributions).
    """
    def _mat_process(M):
        gamma = np.sum(M, axis=0).reshape(-1, 1)
        n = len(M)
        return n, gamma, gamma*(n-1)*gamma.T
    na, alpha, lambda_a = _mat_process(A)
    nb, beta, lambda_b = _mat_process(B)
    return lstsq(lambda_a+lambda_b, na*nb*(alpha-beta))[0].flatten()

__all__ = [
    'split_svm',
    'split_median',
    'split_mean',
    'split_dist'
]